/*
 * $RCSfile: libsdict.c,v $
 * $Author: swaj $
 * $Revision: 1.17 $  
 */

/*
 * Was written by Alexey Semenoff (c) 2001-2006
 * See COPYING for copyiright policy
 */


#define _GNU_SOURCE
#define SDICT_LIBRARY_SO

#ifdef _WIN32
#pragma warning(disable :4013)
#pragma warning(disable :4018)
#include "win32.h"
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifndef _WIN32
#include <stdint.h>
#include <unistd.h>
#endif //_WIN32

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "zlib/unzlib.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(DEBUG) && !defined(DEBUG_LIBSDICT)
#undef DEBUG
#endif

//#define DEBUG 1

#include "defines.h"
#include "debugging.h"
#include "util.h"
#include "libunicode.h"
#define __LIBSDICT_C 1
#include "libsdict_int.h"
#undef __LIBSDICT_C

#ifdef _WIN32
#define S_OPEN  _open
#define S_CLOSE _close
#define S_LSEEK _lseek
#define S_READ  _read
#else //  UNIX
#define S_OPEN  open
#define S_CLOSE close
#define S_LSEEK lseek
#define S_READ  read
#define O_BINARY 0
#endif // _WIN32


#ifdef _WIN32
#include <windows.h>
#include "config.h"

BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved )
{
	/*
	switch (ul_reason_for_call) {
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
	*/
	return TRUE;
}
#endif // _WIN32



/*
 * Public interface
 */


/*
 * Version information
 */
SDICT_API const char* sdict_get_library_version (void)
{
	return (VERSION);
}

/*
 * Creates new dictionary object
 */
SDICT_API Sdict* sdict_new (int word_max, int art_max)
{
	Sdict *sdict;

	DBG("sdict_new()");


	unless ((sdict = smalloc(sizeof(Sdict)))) {
		DBG("smalloc() error");
		return NULL;
	}
	sdict->fd = -1;

	sdict->word_max = word_max > 0 ? word_max : WORD_MAX_DFL;
	sdict->art_max = art_max > 0 ? art_max : ART_MAX_DFL;

/* err_str */
	unless ( (sdict->err_str = smalloc(ERROR_STRING_MAX + 1)) ) {
		DBG("smalloc() error");
		sfree (sdict);
		return NULL;
	}

/* last_word */
	unless ( (sdict->last_word = smalloc(sdict->word_max + 1)) ) {
		DBG("smalloc() error");
		sfree (sdict->err_str);
		sfree (sdict);
		return NULL;
	}

/* last_art */
	unless ( (sdict->last_art = smalloc(sdict->art_max + 1)) ) {
		DBG("smalloc() error");
		sfree (sdict->err_str);
		sfree (sdict->last_word);
		sfree (sdict);
		return NULL;
	}

/* raw_buf */
	unless ( (sdict->raw_buf = smalloc(sdict->art_max + 1)) ) {
		DBG("smalloc() error");
		sfree (sdict->err_str);
		sfree (sdict->last_word);
		sfree (sdict->last_art);
		sfree (sdict);
		return NULL;
	}

	return sdict;
}


/*
 * Deletes dictionary
 */
SDICT_API void sdict_del (Sdict *sdict)
{
	DBG("sdict_del()");

	unless (sdict) {
		DBG("null argument");
		return;
	}

	sfree (sdict->err_str);
	sfree (sdict->last_word);
	sfree (sdict->last_art);
	sfree (sdict->raw_buf);

	sfree (sdict);

	return;
}


/*
 *  Loads .dct file
 */
SDICT_API int sdict_load_dictionary (Sdict* sdict, const char* filename)
{
	int fd;

	DBG("sdict_load_dictionary()");

	unless (sdict && filename && strlen(filename)) {
		DBG("null arg");
		return RET_FAILED;
	}

	DBG("loading file '%s'", filename);
	if ((fd= S_OPEN (filename, O_RDONLY | O_BINARY ))<=0) {
		DBG("unable open file '%s' :%s", filename, strerror(errno));

#define SAVE_ERROR(__a) strncpy(sdict->err_str, (__a), ERROR_STRING_MAX)		
#define CLOSE_DCT_FD S_CLOSE  (sdict->fd); sdict->fd = -1

		SAVE_ERROR(ERR_CANT_OPEN);
		return RET_FAILED;
	}

	return sdict_load_dictionary_fd (sdict, fd);
}

#ifdef _WIN32
SDICT_API int sdict_load_dictionary_w (Sdict* sdict, const wchar_t* filename)
{
	int fd;

	DBG("sdict_load_dictionary()");

	unless (sdict && filename) {
		DBG("null arg");
		return RET_FAILED;
	}

	if ((fd = _wopen (filename, O_RDONLY | O_BINARY ))<=0) {
		DBG("unable open file, reason :%s", filename, strerror(errno));
		SAVE_ERROR(ERR_CANT_OPEN);
		return RET_FAILED;
	}

	return sdict_load_dictionary_fd (sdict, fd);
}
#endif // WIN32


#ifndef _WIN32
SDICT_API int sdict_load_dictionary_file (Sdict* sdict, FILE* file)
{
	int fd;

	DBG("sdict_load_dictionary_file()");
	
	unless (sdict && file) {
		DBG("null arg");
		return RET_FAILED;
	}

	fd = fileno (file);
	return (fd <= 0) ? RET_FAILED : sdict_load_dictionary_fd (sdict, fd);
}
#endif // _WIN32

static
int sdict_load_dictionary_fd (Sdict* sdict, int fd)
{
	Sdct_hdr buf;
	uint8_t buff[DCT_HDR_SIZE];
	Unit *un;

	DBG("sdict_load_dictionary()");


        DBG("reading %d bytes header from fd '%d'", DCT_HDR_SIZE, fd);
/* 	if ((read(fd, &buf, sizeof(Sdct_hdr)))!=sizeof(Sdct_hdr)) { */
	if ( ( S_READ (fd, buff, DCT_HDR_SIZE) != DCT_HDR_SIZE) ) {

		DBG("unable to read header");
		SAVE_ERROR(ERR_UNABLE_READ_HDR);
		S_CLOSE (fd);
		return RET_FAILED;
	}

/*	if (memcmp(&buf.sig, SDCT_SIG, 4)) { */
	if (memcmp(buff, SDCT_SIG, 4)) {
		DBG("wrong signature 0x%x 0x%x 0x%x 0x%x ", buff[0], buff[1], buff[2], buff[3] );
		SAVE_ERROR(ERR_WRONG_SIG);
		S_CLOSE (fd);
		return RET_FAILED;
	}

	sdict->fd = fd;

	
	memset (sdict->w_lang, 4, 0x0);
	memcpy (sdict->w_lang, &buff[0x4], 3);

	memset (sdict->a_lang, 4, 0x0);
	memcpy (sdict->a_lang, &buff[0x7], 3);

	sdict->comp = buff[0xa];


	memcpy(&sdict->w_total, &buff[0xb], sizeof(uint32_t));
	memcpy(&buf.sindex_len, &buff[0xf], sizeof(uint32_t));
	memcpy(&buf.title_ptr, &buff[0x13], sizeof(uint32_t));
	memcpy(&buf.copyr_ptr, &buff[0x17], sizeof(uint32_t));
	memcpy(&buf.ver_ptr, &buff[0x1b], sizeof(uint32_t));
	memcpy(&buf.sindex_ptr, &buff[0x1f], sizeof(uint32_t));
	memcpy(&buf.findex_ptr, &buff[0x23], sizeof(uint32_t));
	memcpy(&buf.art_ptr, &buff[0x27], sizeof(uint32_t));

/* title */
	unless ((un=read_unit(sdict, buf.title_ptr))) {
		DBG("unable read title information");
		SAVE_ERROR(ERR_UNABLE_READ_TIT);
		CLOSE_DCT_FD;
		return RET_FAILED;
	}

	sdict->title = string_from_unit(un);
	del_unit(un);

	unless (sdict->title) {
		DBG("unable get string");
		SAVE_ERROR(ERR_UNABLE_READ_TIT);

		CLOSE_DCT_FD;
		return RET_FAILED;
	}

/* copyright */
	unless ((un=read_unit(sdict, buf.copyr_ptr))) {
		DBG("unable read copyright information");
		SAVE_ERROR(ERR_UNABLE_READ_COPY);

		sfree(sdict->title);

		CLOSE_DCT_FD;
		return RET_FAILED;
	}

	sdict->copy = string_from_unit(un);
	del_unit(un);

	unless (sdict->copy) {
		DBG("unable get string");
		SAVE_ERROR(ERR_UNABLE_READ_COPY);
		sfree(sdict->title);


		CLOSE_DCT_FD;
		return RET_FAILED;
	}


/* version */
	unless ((un=read_unit(sdict, buf.ver_ptr))) {
		DBG("unable read version information");
		SAVE_ERROR(ERR_UNABLE_READ_VER);
				
		sfree(sdict->title);
		sfree(sdict->copy);

		CLOSE_DCT_FD;
		return RET_FAILED;
	}

	sdict->ver = string_from_unit(un);
	del_unit(un);

	unless (sdict->ver) {
		DBG("unable get string");
		SAVE_ERROR(ERR_UNABLE_READ_VER);
				
		sfree(sdict->title);
		sfree(sdict->copy);

		CLOSE_DCT_FD;
		return RET_FAILED;
	}


/* short index */
	sdict->sindex_len = buf.sindex_len;
	unless((sdict->sindex = read_sindex(sdict, buf.sindex_ptr))) {
		DBG("short index read failed");
		SAVE_ERROR(ERR_UNABLE_READ_SINDEX);

		sfree(sdict->title);
		sfree(sdict->copy);
		sfree(sdict->ver);

		CLOSE_DCT_FD;
		return RET_FAILED;
	}		

	
	if (parse_sindex(sdict)!=RET_SUCCESS) {
		DBG("short index parse failed");
		SAVE_ERROR(ERR_UNABLE_PARSE_SINDEX);

		sfree(sdict->title);
		sfree(sdict->copy);
		sfree(sdict->ver);
		sfree(sdict->sindex);
		sdict->sindex_len = 0;

		CLOSE_DCT_FD;
		return RET_FAILED;
	}		


	sdict->findex = buf.findex_ptr;
	sdict->art = buf.art_ptr;

	sdict->findex_pos = sdict->findex;


        DBG("*** Dictionary info:");
        DBG("    w_lang    : '%s'", sdict->w_lang);
        DBG("    a_lang    : '%s'", sdict->a_lang);
        DBG("    comp      : '%c'", sdict->comp);
        DBG("    w_total   : '%d'", sdict->w_total);
        DBG("    title     : '%s'", sdict->title);
        DBG("    copyright : '%s'", sdict->copy);
        DBG("    version   : '%s'", sdict->ver);
        DBG(" ");
        DBG("    fndx off  : '0x%x'", sdict->findex);
        DBG("    art  off  : '0x%x'", sdict->art);

	return RET_SUCCESS;
}


/*
 * unload dct file 
 */
SDICT_API void sdict_unload_dictionary (Sdict *sdict)
{
	DBG("sdict_unload_dictionary()");

	unless (sdict) {
		DBG("null argument");
		return;
	}

	sfree(sdict->title);
	sfree(sdict->copy);
	sfree(sdict->ver);
	sfree(sdict->sindex);
	sdict->sindex_len = 0;

	CLOSE_DCT_FD;

	sdict->findex = 0;
	sdict->art = 0;

	sdict->findex_pos = 0;
}


/*
 * read current word, rewind to next pos, fill sdict->last_word sdict->last_art
 */
SDICT_API int sdict_get_next_word(Sdict *sdict)
{
	uint16_t next, prev;
	uint32_t aptr;
	int word_len;


	DBG("sdict_get_next_word()");


	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return RET_FAILED;
	}


	unless (sdict->findex_pos)
		sdict->findex_pos = sdict->findex;

	DBG("sdict->findex_pos = 0x%x", sdict->findex_pos);
	if ( S_LSEEK (sdict->fd, sdict->findex_pos, SEEK_SET) != sdict->findex_pos) {
		DBG("lseek error %s", strerror(errno));
		return RET_FAILED;
	}

        /* get next */
	if ( S_READ (sdict->fd, &next, sizeof(next)) != sizeof(next) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("next = 0x%x", next);
		unless (next) {
			DBG("last word reached");
			return RET_FAILED;
		}
	}

	/* get prev */
	if ( S_READ (sdict->fd, &prev, sizeof(prev)) != sizeof(prev) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("prev = 0x%x", prev);
	}

	/* get article ptr */
	if ( S_READ (sdict->fd, &aptr, sizeof(aptr)) != sizeof(aptr) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("aptr = 0x%x", aptr);
	}

	/* get word */
	word_len = next - (2 * sizeof(next) + sizeof(aptr));
	
	if (word_len > sdict->word_max ) {
		DBG("word too long");
	}


	if ( S_READ (sdict->fd, sdict->last_word, MIN(sdict->word_max, word_len)) != word_len ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		sdict->last_word[MIN(sdict->word_max, word_len)] = 0x0;
		DBG("word = %s", sdict->last_word);
	}
	

	sdict->last_art_ptr = (pointer)aptr;
	sdict->findex_pos += next;

	return RET_SUCCESS;
}


/*
 * read word from previous position, rewind 1 word , fill sdict->last_word sdict->last_art
 */
SDICT_API int sdict_get_prev_word(Sdict *sdict)
{
	uint16_t next, prev;
	uint32_t aptr;
	int word_len;
	int noff;

	DBG("sdict_get_prev_word()");

	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return RET_FAILED;
	}


	unless (sdict->findex_pos)
		sdict->findex_pos = sdict->findex;

	DBG("sdict->findex_pos = 0x%x", sdict->findex_pos);
	if ( S_LSEEK (sdict->fd, sdict->findex_pos, SEEK_SET) != sdict->findex_pos) {
		DBG("lseek error %s", strerror(errno));
		return RET_FAILED;
	}

        /* get next */
	if ( S_READ (sdict->fd, &next, sizeof(next)) != sizeof(next) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("next = 0x%x", next);
	}

	/* get prev */
	if ( S_READ (sdict->fd, &prev, sizeof(prev)) != sizeof(prev) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("prev = 0x%x", prev);
		unless (prev) {
			DBG("first word reached");
			return RET_FAILED;
		}
	}

	noff = sdict->findex_pos - prev;

	DBG("noff  = 0x%x", noff);
	if ( S_LSEEK (sdict->fd, noff, SEEK_SET) != noff) {
		DBG("lseek error %s", strerror(errno));
		return RET_FAILED;
	}


        /* get next */
	if ( S_READ (sdict->fd, &next, sizeof(next)) != sizeof(next) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("next = 0x%x", next);
	}


	/* get prev */
	if ( S_READ (sdict->fd, &prev, sizeof(prev)) != sizeof(prev) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("prev = 0x%x", prev);
	}

	/* get article ptr */
	if ( S_READ (sdict->fd, &aptr, sizeof(aptr)) != sizeof(aptr) ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		DBG("aptr = 0x%x", aptr);
	}

	/* get word */
	word_len = next - (2 * sizeof(next) + sizeof(aptr));
	
	if (word_len > sdict->word_max ) {
		DBG("word too long");
	}

	if ( S_READ (sdict->fd, sdict->last_word, MIN(sdict->word_max, word_len)) != word_len ) {
		DBG("read failed");
		return RET_FAILED;
	} else {
		sdict->last_word[MIN(sdict->word_max, word_len)] = 0x0;
		DBG("word = %s", sdict->last_word);
	}
	
	sdict->last_art_ptr = (pointer)aptr;
	sdict->findex_pos = noff;

	return RET_SUCCESS;
}


/*
 * unpack article under  Sdict->last_art_ptr, put into Sdict->last_art as utf8
 */
SDICT_API int sdict_get_article (Sdict *sdict)
{
	DBG("sdict_get_article()");


	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return RET_FAILED;
	}

	return read_article_unit (sdict, (int)sdict->last_art_ptr);
}


/* 
 * Find word in dictionary, word is utf8 string
 */
SDICT_API int sdict_find_word (Sdict *sdict, utf8 *word)
{
	utf32 *ptr, *ptrt;
	int len, i, a;
	Sindex *sindex;
	uint32_t findex_pos_tmp;
	int lenn, slenn;

	DBG("sdict_find_word()");


	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return RET_FAILED;
	}

	unless (word) {
		DBG("null argument");
		return RET_FAILED;
	}

	len = utf8_strlen(word);

	DBG("find word '%s', len = %d", word, len);

	unless(len) {
		DBG("word zero length");
		return RET_FAILED;
	}

	unless ((ptr = utf8_to_utf32 (word))) {
		return RET_FAILED;
	}

	unless ((ptrt = utf8_to_utf32 (word))) {
		return RET_FAILED;
	}


	switch (len) {

	case 1:
		DBG("1 char");
		sindex = sdict->sindex_1;
		lenn = 1;
		slenn = sdict->sindex_1_len;
		break;

	case 2:
		DBG("2 chars");
		sindex = sdict->sindex_2;
		lenn = 2;
		slenn = sdict->sindex_2_len;
		break;

	default:
		DBG("3 chars or mode ");
		sindex = sdict->sindex_3;
		lenn = 3;
		slenn = sdict->sindex_3_len;
	}


	if (slenn < LOOK_THRU_WORDS_ANYWAY) {
                int sl = (sdict->w_total < LOOK_THRU_WORDS_ANYWAY) ?
                        sdict->w_total : LOOK_THRU_WORDS_ANYWAY ;
                DBG ("slenn TOO small, instead use %d", sl);
                slenn = sl;
        }

	for (i=0; i < slenn; i++) { /* look through sindex   */

		noDBG(" '%c%c%c' : 0x%x", (uchar)sindex->item[0], (uchar)sindex->item[1], (uchar)sindex->item[2], sindex->ptr);

		//!! check boundary
		if ( ! *( (uint32_t*) sindex->item) ) {
			DBG ("OOPS, out of sindex !");
                        break;
		}
		//!!

		if (utf32_strncmp (ptr, sindex->item, lenn)==RET_SUCCESS) {
			//!! search to this position - it is next
			int max_file_pos=0;
			//!!

			DBG("found at ptr 0x%x", sindex->ptr + sdict->findex );
			DBG(" '%c%c%c' : 0x%x", (uchar)sindex->item[0], (uchar)sindex->item[1], (uchar)sindex->item[2], sindex->ptr);
				
			findex_pos_tmp = sdict->findex_pos;
			sdict->findex_pos = sindex->ptr + sdict->findex; 
			DBG("set sdict->findex_pos to 0x%x ", sdict->findex_pos);
			
			//!! get next fpos from sindex and later search to 
			sindex++;
			if ( *( (uint32_t*) sindex->item) && sindex->ptr) {
				if ( sindex->ptr + sdict->findex > sdict->findex_pos) {
					max_file_pos = sindex->ptr + sdict->findex;
				}
			}
			sindex--;
			DBG("max_file_pos= 0x%x", max_file_pos );
			//!!

			a = slenn;
			//DBG("slenn = %d", slenn);
			

			while (1 /*a-- */) {

				//DBG("1> sdict->findex_pos to 0x%x ", sdict->findex_pos);

				if (sdict_get_next_word(sdict) ==RET_FAILED) {
					DBG("sdict_get_next_word() failed");
					sfree(ptr); sfree(ptrt);
					return RET_FAILED;
				}

				//DBG("2> sdict->findex_pos to 0x%x ", sdict->findex_pos);
				//return RET_FAILED;

				//!! check if we are on the next sindex
				if (max_file_pos && sdict->findex_pos > max_file_pos) {
					DBG ("NEXT sindex, stop");
					sfree(ptr); sfree(ptrt);
					return RET_FAILED;
				}
				//!!


				if ( (strlen(word)==strlen(sdict->last_word)) &&
				     (strcmp(word,sdict->last_word) == 0) ) {
					DBG("exact match");
					sdict->findex_pos = findex_pos_tmp;
					sfree(ptr); sfree(ptrt);
					return sdict_get_article(sdict);
				}


				noDBG("index found word = '%c%c%c' sindex->item = '%c%c%c'",
				    (uchar)ptr[0],
				    (uchar)ptr[1],
				    (uchar)ptr[2],
				    (uchar)sindex->item[0],
				    (uchar)sindex->item[1],
				    (uchar)sindex->item[2] );

				utf8_to_utf32_ninplace (sdict->last_word, ptrt, lenn);
				
				if (utf32_strncmp (ptrt, sindex->item, lenn)==RET_FAILED) {

					DBG("next index found word = '%c%c%c' sindex->item = '%c%c%c'",
					    (uchar)ptrt[0],
					    (uchar)ptrt[1],
					    (uchar)ptrt[2],
					    (uchar)sindex->item[0],
					    (uchar)sindex->item[1],
					    (uchar)sindex->item[2] );


					sdict->findex_pos = findex_pos_tmp;
					sfree(ptr);
					sfree(ptrt);
					return RET_FAILED;
	
				}


				DBG("next in search");

			} //\\while ...

			sfree(ptr);
			sfree(ptrt);

			DBG("return failed");

			return RET_FAILED;
		} //\\ if (utf32_strncmp (ptr, sindex->item, lenn)==RET_SUCCESS) 

		sindex++;
	} //\\ for (i=0; i < slenn; i++) {  look through sindex 

	sfree(ptr);
	sfree(ptrt);
	return RET_FAILED;

}


/* 
 * Move full position index according to word, word is utf8 string
 */
SDICT_API int sdict_seek_word (Sdict *sdict, utf8 *word)
{
	utf32 *ptr, *ptrt;
	int len, i, a;
	Sindex *sindex;
	uint32_t findex_pos_tmp;
	int lenn, slenn;

	DBG("sdict_seek_word()");


	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return RET_FAILED;
	}

	unless (word) {
		DBG("null argument");
		return RET_FAILED;
	}

	len = utf8_strlen(word);

	DBG("find word '%s', len = %d", word, len);

	unless(len) {
		DBG("word zero length");
		return RET_FAILED;
	}

	unless ((ptr = utf8_to_utf32 (word))) {
		return RET_FAILED;
	}

	unless ((ptrt = utf8_to_utf32 (word))) {
		return RET_FAILED;
	}


	switch (len) {

	case 1:
		DBG("1 char");
		sindex = sdict->sindex_1;
		lenn = 1;
		slenn = sdict->sindex_1_len;
		break;

	case 2:
		DBG("2 chars");
		sindex = sdict->sindex_2;
		lenn = 2;
		slenn = sdict->sindex_2_len;
		break;

	default:
		DBG("3 chars or mode ");
		sindex = sdict->sindex_3;
		lenn = 3;
		slenn = sdict->sindex_3_len;
	}

	if (slenn < LOOK_THRU_WORDS_ANYWAY) {
                int sl = (sdict->w_total < LOOK_THRU_WORDS_ANYWAY) ?
                        sdict->w_total : LOOK_THRU_WORDS_ANYWAY ;
                DBG ("slenn TOO small, instead use %d", sl);
                slenn = sl;
        }

	if (len <=3 )
	{
		for (i=0; i < slenn; i++) { /* look through sindex */
			//!! check boundary
			if ( ! *( (uint32_t*) sindex->item) ) {
				DBG ("OOPS, out of sindex !");
				break;
			}
			//!!

			noDBG(" '%c%c%c' : 0x%x",
			      (uchar)sindex->item[0],
			      (uchar)sindex->item[1],
			      (uchar)sindex->item[2],
			      sindex->ptr
				);

			if (utf32_strncmp (ptr, sindex->item, lenn)==RET_SUCCESS) {
				DBG("found at ptr 0x%x",
				    sindex->ptr + sdict->findex );
				DBG(" '%c%c%c' : 0x%x",
				    (uchar)sindex->item[0],
				    (uchar)sindex->item[1],
				    (uchar)sindex->item[2], sindex->ptr
					);

				sdict->findex_pos = sindex->ptr + sdict->findex; 
				sdict->prev_search_pos = sdict->findex_pos;

				sfree(ptr);
				sfree(ptrt);

				//sdict_printd ("*** Hit! ***");
				return RET_SUCCESS;
			}
		
			sindex++;
		} //\\for (i=0; i < slenn;
		
		sfree(ptr);
		sfree(ptrt);
		//sdict_printd ("*** Miss! ***");
		sdict->prev_search_pos = 0;

		return RET_FAILED;

	} else	{ // ---------------    > 3 
		findex_pos_tmp = sdict->findex_pos;

		for (i=0; i < slenn; i++) { /* look through sindex   */

			//!! check boundary
			if ( ! *( (uint32_t*) sindex->item) ) {
				DBG ("OOPS, out of sindex !");
				break;
			}
			//!!

			noDBG(" '%c%c%c' : 0x%x",
			      (uchar)sindex->item[0],
			      (uchar)sindex->item[1],
			      (uchar)sindex->item[2],
			      sindex->ptr
				);

			if (utf32_strncmp (ptr, sindex->item, lenn)==RET_SUCCESS) {
				//!! search to this position - it is next
				int max_file_pos=0;
				//!!
				
			//DBG("found at ptr 0x%x", sindex->ptr + sdict->findex );
			//DBG(" '%c%c%c' : 0x%x", (uchar)sindex->item[0], (uchar)sindex->item[1], (uchar)sindex->item[2], sindex->ptr);
				
				//findex_pos_tmp = sdict->findex_pos;

				if (sdict->prev_search_pos ) {
					sdict->findex_pos = sdict->prev_search_pos;
					DBG ("HIT!, continue search");
					
				} else {
					sdict->findex_pos = sindex->ptr + sdict->findex; 

					DBG ("HIT!, base search");
				}

				//!! get next fpos from sindex and later search to 
				sindex++;
				if ( *( (uint32_t*) sindex->item) && sindex->ptr) {
					if ( sindex->ptr + sdict->findex > sdict->findex_pos) {
						max_file_pos = sindex->ptr + sdict->findex;
					}
				}
				sindex--;
				DBG("max_file_pos= 0x%x", max_file_pos );
				//!!
		

				a = slenn;

				while (1 /*a--*/) {
					if (sdict_get_next_word(sdict) ==RET_FAILED) {
						sfree(ptr);
						sfree(ptrt);
						return RET_FAILED;
					}

					//!! check if we are on the next sindex
					if (max_file_pos && sdict->findex_pos > max_file_pos) {
						DBG ("NEXT sindex, stop");
						sfree(ptr);
						sfree(ptrt);
						return RET_FAILED;
					}
					//!!

					if ( utf8_strncmp( word, sdict->last_word, len ) == 0) {
						sfree(ptr);
						sfree(ptrt);

						sdict_get_prev_word(sdict); // Always nostrip !!!
						sdict->prev_search_pos = sdict->findex_pos;
						return RET_SUCCESS; //return sdict_get_article(sdict);
					}


					utf8_to_utf32_ninplace (sdict->last_word, ptrt, lenn);
				
					if (utf32_strncmp (ptrt, sindex->item, lenn)==RET_FAILED) {
						sdict->findex_pos = findex_pos_tmp;
						sfree(ptr);
						sfree(ptrt);

						DBG ("failed, out of sindex");
						return RET_FAILED;
					}

				//DBG("next in search");
				}

				sfree(ptr);
				sfree(ptrt);

				DBG ("failed, out of a--");
				return RET_FAILED;		

			} //\\ (utf32_strncmp (pt
		
			sindex++;

		} //\\ (i=0; i < slenn; i++
	
		sfree(ptr);
		sfree(ptrt);
		DBG("failed, out of global sindex search");
		return RET_FAILED;
	}
}


/**** 
 *  *
 *  *
 ****/


/*
 * Private interface
 */


/*
 * malloc() then copy data from unit
 */
static
uchar* string_from_unit(Unit *u)
{
	uchar *p;

	unless (u) {
		DBG("null pointer");
		return NULL;
	}

	unless (u->len) {
		DBG("null len");
		return NULL;
	}

	if (u->len >=UNIT_MAX_LEN) {
		DBG("too big");
		return NULL;
	}

	unless ((p=smalloc(u->len+1))) {
		DBG("malloc() error");
		return NULL;
	}

	strncpy(p, &u->data, u->len);
	return p;
}


/*
 * free unit memory
 */
static
void del_unit(Unit *u)
{
	if (u) sfree(u);
}


/*
 * load short index and convert utf32 -> utf16
 */
static
Sindex* read_sindex(Sdict *sdict, int offset)
{
	Sindex *ndx1, *ndx2, *ndx3;
	uLongf dl;
	int i, rv, t;


	unless (sdict) {
		DBG("null pointer");
		return NULL;
	}

	if (sdict->fd <=0) {
		DBG("bad fd");
		return NULL;
	}

	if ( S_LSEEK (sdict->fd, offset, SEEK_SET) != offset) {
		DBG("lseek error %s", strerror(errno));
		return NULL;
	}

	unless ((ndx1=smalloc(sdict->sindex_len * sizeof(Sindex) + ZLIB_HEADER ))) {
		DBG("malloc() failed");
		return NULL;
	}

	if ( S_READ (sdict->fd, ndx1,
		 sdict->sindex_len * sizeof(Sindex) + ZLIB_HEADER ) != sdict->sindex_len * sizeof(Sindex) + ZLIB_HEADER ) {
		DBG("unable to read sindex");
		sfree(ndx1);
		return NULL;
	}

	DBG("read ok, %d bytes", sdict->sindex_len * sizeof(Sindex) + ZLIB_HEADER ); 


	switch (sdict->comp) {

	case '0':
		DBG("no compression needed");
		//!!
		unless ((ndx3=smalloc( (sdict->sindex_len+1) * sizeof(Sindex)))) {
			DBG("malloc() failed");
			sfree(ndx1);
			return NULL;
		}

		for (i=0; i < sdict->sindex_len; i++) {
			ndx3[i].ptr = ndx1[i].ptr;
			
			for (t=0; t<3; t++) {
				ndx3[i].item[t] = ndx1[i].item[t];
			}
		}

		sfree(ndx1);
		return ndx3;

	case '1':
		DBG("zlib compresssion");

		//!!
		unless ((ndx2=smalloc( (sdict->sindex_len+1) * sizeof(Sindex)))) {
			DBG("malloc() failed");
			sfree(ndx1);
			return NULL;
		}

		dl = sdict->sindex_len * sizeof(Sindex);

		if ((rv=zuncompress ( (Bytef*)ndx2,
				     &dl,
				     (Bytef*)ndx1,
				     (uLong)(sdict->sindex_len * sizeof(Sindex))) )) {

			DBG("zuncompress failed, rv = %d", rv);
			sfree(ndx1);
			sfree(ndx2);
			return NULL;
		}
		sfree(ndx1);

		DBG("decompressed %d bytes", (int)dl);

		unless (dl) {
			DBG("decompression error");
			sfree(ndx2);
			return NULL;
		}

		//!!
		unless ((ndx3=smalloc((sdict->sindex_len+1) * sizeof(Sindex)))) {
			DBG("malloc() failed");
			sfree(ndx1);
			return NULL;
		}

		for (i=0; i < sdict->sindex_len; i++) {
			ndx3[i].ptr = ndx2[i].ptr;
			for (t=0; t<3; t++) {
				ndx3[i].item[t] = ndx2[i].item[t];
			}
		}

		sfree(ndx2);
		return ndx3;

	default:
		DBG("unsupported compression");
	}

	return NULL;
}


/*
 * parse short index, fill 1st, 2nd, 3rd subindices as well as their lengths
 */
static
int parse_sindex(Sdict *sdict)
{
	int i;
	Sindex *ptr;

	unless (sdict) {
		DBG("null args");
		return RET_FAILED;
	}

	ptr = sdict->sindex;


	sdict->sindex_1 = ptr;


	sdict->sindex_1_len = 0;
	sdict->sindex_2_len = 0;
	sdict->sindex_3_len = 0;

	for (i=0; i< sdict->sindex_len; i++) {
		if (utf32_strnlen(ptr->item, 3) > 1) break;
		sdict->sindex_1_len++;
		ptr++;
		//DBG("%c", (uchar)ptr->item[0]);
		//DBG("0x%x", ptr->item[0]);
	}


	sdict->sindex_2 = ptr;

	for (;i< sdict->sindex_len; i++) {
		if (utf32_strnlen(ptr->item, 3) > 2) break;
		sdict->sindex_2_len++;
		ptr++;
		//DBG("%c%c", (uchar)ptr->item[0], (uchar)ptr->item[1]);
		//DBG("0x%x 0x%x", ptr->item[0], ptr->item[1]);
	}

	sdict->sindex_3 = ptr;

	for (;i< sdict->sindex_len; i++) {
		sdict->sindex_3_len++;
		ptr++;
		//DBG("%c%c%c", (uchar)ptr->item[0], (uchar)ptr->item[1], (uchar)ptr->item[2] );
		//DBG("0x%x 0x%x 0x%x", ptr->item[0], ptr->item[1], ptr->item[2] );
	}

	DBG ("parse_sindex(), sdict->sindex_len= %d; sdict->sindex_1_len= %d; sdict->sindex_2_len= %d; sdict->sindex_3_len= %d", sdict->sindex_len, sdict->sindex_1_len, sdict->sindex_2_len, sdict->sindex_3_len);

	return RET_SUCCESS;
}


/*
 * general read unit malloc() INFLATE_ZLIB_MAX_RATIO x size,
 * seek file offset, decompress if needed 
 */
static
Unit* read_unit (Sdict *sdict, int offset)
{
	uint32_t len;
	Unit *unit, *unit_d;
	int rv;


	unless (sdict) {
		DBG("null pointer");
		return NULL;
	}

	if (sdict->fd <= 0) {
		DBG("bad fd");
		return NULL;
	}

	if ( S_LSEEK (sdict->fd, offset, SEEK_SET) != offset) {
		DBG("lseek error %s", strerror(errno));
		return NULL;
	}

	if ( S_READ (sdict->fd, &len, sizeof(len)) != sizeof(len) ) {
		DBG("unable to read header");
		return NULL;
	}

	if (len >= UNIT_MAX_LEN) {
		DBG("unit too big");
		return NULL;
	}

	DBG("unit len %d", len);

	unless ((unit=smalloc(len + sizeof(uint32_t)))) {
		DBG("malloc() failed");
		return NULL;
	}

	if ( S_READ (sdict->fd, &unit->data, len) != len) {
		DBG("unable to read from file: %s", strerror(errno));
		sfree(unit);
		return NULL;
	}

	unit->len=len;

	switch (sdict->comp) {
	case '0':
		DBG("no decompression needed");
		return unit;

	case '1':
		DBG("zlib compression");
		unless ((unit_d=smalloc(len * INFLATE_ZLIB_MAX_RATIO + sizeof(uint32_t)))) {
			DBG("malloc() failed");
			sfree(unit);
			return NULL;
		}

		unit_d->len = len * INFLATE_ZLIB_MAX_RATIO;

		if ((rv=zuncompress (&unit_d->data, (uLongf *)&unit_d->len, &unit->data, (uLong)unit->len))) {
			DBG("uncompress failed, rv = %d", rv);
			sfree(unit);
			sfree(unit_d);
			return NULL;
		}
		
		DBG("size: %d;  data: >%s<", unit_d->len, &unit_d->data);
		sfree(unit);
		return unit_d;

	default:
		DBG("unsupported compression");
		sfree(unit);
	}
	return NULL;
}


/*
 * read article unit with no malloc() sdict->last_art/sdict->last_art_raw are used
 * seek file offset, decompress if needed 
 */
static
int read_article_unit (Sdict *sdict, int offset)
{
	uint32_t len;
	int rv;
	uLongf dlen;

	unless (sdict) {
		DBG("null pointer");
		return RET_FAILED;
	}

	if (sdict->fd <= 0) {
		DBG("bad fd");
		return RET_FAILED;
	}

	if ( S_LSEEK (sdict->fd, offset + sdict->art, SEEK_SET) != offset + sdict->art) {
		DBG("lseek error %s", strerror(errno));
		return RET_FAILED;
	}

	if ( S_READ (sdict->fd, &len, sizeof(len)) != sizeof(len) ) {
		DBG("unable to read header");
		return RET_FAILED;
	}

	if (len >= sdict->art_max ) {
		DBG("unit too big, truncating");
		len = sdict->art_max;
	}

	DBG("unit len %d", len);

	if ( S_READ (sdict->fd, sdict->raw_buf, len) <= 0) {
		DBG("unable to read from file: %s", strerror(errno));
		return RET_FAILED;
	}

	switch (sdict->comp) {

	case '0':
		DBG("no decompression needed");
		memcpy(sdict->last_art, sdict->raw_buf, len);
		sdict->last_art[len] = 0x0;
		return RET_SUCCESS;

	case '1':
		DBG("zlib compression");

		dlen = sdict->art_max;

		rv=zuncompress (sdict->last_art, &dlen, sdict->raw_buf, len);

		if ( (rv != Z_OK) && (rv != Z_BUF_ERROR)) {

			DBG("uncompress failed, rv = %d", rv);
			return RET_FAILED;
		}
		
		DBG("size: %d", dlen);
		
		sdict->last_art[dlen] = 0x0;
		return RET_SUCCESS;

	default:
		DBG("unsupported compression");
	}
	return RET_FAILED;
}


/*
 * Rewind full index 
 */
SDICT_API int sdict_rewind (Sdict *sdict)
{
	DBG("sdict_rewind()");

	unless (sdict) {
		DBG("null argument");
		return RET_FAILED;
	}

	sdict->findex_pos = sdict->findex;

	return RET_SUCCESS;
}


/*
 * Seek to approx position in dictionary 
 *  0    - start
 *  500  - ~50%
 *  1000 - end
 */
SDICT_API int sdict_seek_to (Sdict *sdict, int pos) // 0 - start, 1000 - end, 500 - 50%
{
	Sindex *sindex;
	int slen, newpos;

	DBG("sdict_seek_to()");

	if (!sdict || pos < 0 || pos > SEEK_POS_MAX ) {
		DBG("wrong argument");
		return RET_FAILED;
	}

	sindex = sdict->sindex_3;
	slen = sdict->sindex_3_len;

	switch (pos) {

	case 0:
		return sdict_rewind (sdict);

	case SEEK_POS_MAX:
		sindex +=slen;
		sindex --;
		sdict->findex_pos = sdict->findex + sindex->ptr;
		break;

	default:
		newpos = (int)( ( (double)pos * (double)slen ) / (double)SEEK_POS_MAX);
		if (newpos >= slen)
			newpos--;

		sindex += newpos;

		sdict->findex_pos = sdict->findex + sindex->ptr;
		break;
	}

	return RET_SUCCESS;
}


SDICT_API int sdict_get_position (Sdict *sdict)
{
	int fsz, pos;

	if (!sdict ) {
		DBG("wrong argument");
		return -1;
	}

	fsz = sdict->art - sdict->findex ;
	if (fsz <= 0)
		fsz = 1;
		
	pos = (int) ( ((double)(sdict->findex_pos-sdict->findex) * (double)SEEK_POS_MAX) / (double)fsz);

	return ( pos > SEEK_POS_MAX ) ? SEEK_POS_MAX : pos;
}


SDICT_API const utf8* sdict_get_dic_title (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->title;
}


SDICT_API const utf8* sdict_get_dic_copyright (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->copy;
}


SDICT_API const utf8* sdict_get_dic_version (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->ver;
}


SDICT_API const utf8* sdict_get_dic_word_lang (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->w_lang;
}


SDICT_API const utf8* sdict_get_dic_article_lang (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->a_lang;
}


SDICT_API int sdict_get_dic_words (Sdict *sdict)
{
	if (!sdict )
		return 0;
	return  sdict->w_total;
}


SDICT_API const char* sdict_get_strerror (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->err_str;
}


SDICT_API const utf8* sdict_get_last_word (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->last_word;
}


SDICT_API const utf8* sdict_get_last_article (Sdict *sdict)
{
	if (!sdict )
		return "";
	return sdict->last_art;
}


SDICT_API unsigned char sdict_get_dic_compression (Sdict *sdict)
{
	if (!sdict )
		return 0;

	return (sdict->comp & 0xf) ;
}


SDICT_API int sdict_get_dic_last_article_ptr (Sdict *sdict)
{
	if (!sdict )
		return -1;

	return (int) sdict->last_art_ptr;
}


SDICT_API int sdict_set_dic_last_article_ptr (Sdict *sdict, int ptr)
{
	if (!sdict )
		return -1;
	
	sdict->last_art_ptr = (void *) ptr;
	return 0;
}




/* __END__ */

