/*
 * $RCSfile: libsdict.h,v $
 * $Author: swaj $
 * $Revision: 1.14 $  
 */

/*
 * Was written by Alexey Semenoff (c) 2001-2006
 * See COPYING for copyiright policy
 */

#ifndef __LIBSDICT_H_
#define __LIBSDICT_H_ 1

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifdef _NOT_DEFINED /* to fool emacs */
}
#endif

#ifdef _WIN32
#ifdef SDICT_LIBRARY_EXPORTS
#define SDICT_API __declspec(dllexport)
#else
#define SDICT_API __declspec(dllimport)
#endif
#else // UNIX
#define SDICT_API
#endif // _WIN32

#ifndef SDICT_LIBRARY_SO
typedef void* Sdict;
#endif //SDICT_LIBRARY_SO

#ifndef __LIBUNICODE_H_
typedef unsigned char utf8;
#endif

/* LIBRARY INTERFACE */

/* version info */
SDICT_API const char* sdict_get_library_version (void);


/* constructor/destructor */
SDICT_API Sdict* sdict_new (int word_max, int art_max);
SDICT_API void  sdict_del (Sdict *sdict);


/* load/unload .dct */
SDICT_API int sdict_load_dictionary (Sdict *sdict, const char* filename);
#ifdef _WIN32
SDICT_API int sdict_load_dictionary_w (Sdict* sdict, const wchar_t* filename);
#endif
#ifndef _WIN32
SDICT_API int sdict_load_dictionary_file (Sdict* sdict, FILE* file);
#endif
SDICT_API void sdict_unload_dictionary (Sdict *sdict);


/* search, seek etc */
SDICT_API int sdict_get_next_word(Sdict *sdic);
SDICT_API int sdict_get_prev_word(Sdict *sdic);
SDICT_API int sdict_get_article (Sdict *sdict);
SDICT_API int sdict_find_word (Sdict *sdict, utf8 *word);
SDICT_API int sdict_seek_word (Sdict *sdict, utf8 *word);
SDICT_API int sdict_rewind (Sdict *sdict);

#define SEEK_POS_MAX 1000                            /* scale used in library */
SDICT_API int sdict_seek_to (Sdict *sdict, int pos); /* i.e  0 -> start; 1000 -> end, 500 -> 50% */
SDICT_API int sdict_get_position (Sdict *sdict);


/* dictionary information */
SDICT_API const utf8* sdict_get_dic_title (Sdict *sdict);
SDICT_API const utf8* sdict_get_dic_copyright (Sdict *sdict);
SDICT_API const utf8* sdict_get_dic_version (Sdict *sdict);
SDICT_API const utf8* sdict_get_dic_word_lang (Sdict *sdict);
SDICT_API const utf8* sdict_get_dic_article_lang (Sdict *sdict);
SDICT_API int sdict_get_dic_words (Sdict *sdict);

SDICT_API unsigned char sdict_get_dic_compression (Sdict *sdict); // 0 -> not compressed, 1 -> zlib
SDICT_API int sdict_get_dic_last_article_ptr (Sdict *sdict);
SDICT_API int sdict_set_dic_last_article_ptr (Sdict *sdict, int ptr);


/* error api */
SDICT_API const char* sdict_get_strerror (Sdict *sdict);


/* word/article after success */
SDICT_API const utf8* sdict_get_last_word (Sdict *sdict);
SDICT_API const utf8* sdict_get_last_article (Sdict *sdict);




#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __LIBSDICT_H_ */



/* __END__ */
