/*
 * $RCSfile: libsdict_int.h,v $
 * $Author: swaj $
 * $Revision: 1.1 $  
 */

/*
 * Was written by Alexey Semenoff (c) 2001-2006
 * See COPYING for copyiright policy
 */

#ifndef __LIBSDICT_INT_H_
#define __LIBSDICT_INTH_ 1


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#ifdef _WIN32

#ifdef SDICT_LIBRARY_EXPORTS
#define SDICT_API __declspec(dllexport)
#else
#define SDICT_API __declspec(dllimport)
#endif

#else // UNIX

#define SDICT_API

#endif // _WIN32


#define ERROR_STRING_MAX 128
#define SDCT_SIG "sdct"
#define UNIT_MAX_LEN 3000000
#define INFLATE_ZLIB_MAX_RATIO 6
#define ZLIB_HEADER 32
#define WORD_MAX_DFL 512
#define ART_MAX_DFL 128000

#define LOOK_THRU_WORDS_ANYWAY 1000

/* Error strings */
#define ERR_CANT_OPEN		"cannot open file"
#define ERR_UNABLE_READ_HDR	"unable to read header"
#define ERR_WRONG_SIG		"wrong signature"
#define ERR_UNABLE_READ_TIT	"unable read title information"
#define ERR_UNABLE_READ_COPY	"unable read copyright information"
#define ERR_UNABLE_READ_VER	"unable read version information"
#define ERR_UNABLE_READ_SINDEX	"short index read failed"
#define ERR_UNABLE_PARSE_SINDEX	"short index parse failed"



/* Information unit uncompressed len, then raw data */
typedef struct {
	uint32_t len;
	uchar data;
} Unit;


/* short index */
typedef struct __Sindex {
	uint32_t item[3];
	uint32_t ptr;
} Sindex;


/* Sdict object */
typedef struct __Sdict {
	int fd;			/* opened file */

	uchar	*err_str;	/* last error */
	int	word_max;	/* max word size */
	int	art_max;	/* max art size */
	
	utf8	*last_word;	/* store word after each  get_next_word() / get_prev_word() */
	utf8	*last_art;	/* store article after read_article(), uncompressed */
	uint8_t	*raw_buf;	/* store article after read_article(), compressed */

	uint8_t w_lang[4];	/* source language */
	uint8_t a_lang[4];	/* dest language */
	uint8_t comp;		/* compression type, only '0' and '1' are supported for now ! */
	uint32_t w_total;	/* words total */

	utf8 *title;		/* Title, */
	utf8 *copy;		/*       copyright */
	utf8 *ver;		/*                 version info (utf8, null-terminated) */

	Sindex *sindex;		/* ptr to full sindex 'a', 'ab', then 'abc' */
	uint32_t sindex_len;	/* length of sindex (units) */
	Sindex *sindex_1;	/* ptr to 'a' sindex */
	uint32_t sindex_1_len;	/* length of sindex 'a' (units) */
	Sindex *sindex_2;	/* ptr to 'ab' sindex */
	uint32_t sindex_2_len;	/* length of sindex 'ab' (units) */
	Sindex *sindex_3;	/* ptr to 'abc' sindex */
	uint32_t sindex_3_len;	/* length of sindex 'abc' (units) */

	uint32_t findex;	/* file offset to full index */
	uint32_t art;		/* file offset to articles */

	uint32_t findex_pos;	/* current full_index position */
	pointer last_art_ptr;	/* store article ptr after each get_next_word() / get_prev_word() */

	uint32_t prev_search_pos;
} Sdict;


typedef struct __Sdct_hdr {

	uint32_t  sindex_len;
	uint32_t  title_ptr;
	uint32_t  copyr_ptr;
	uint32_t  ver_ptr;
	uint32_t  sindex_ptr;
	uint32_t  findex_ptr;
	uint32_t  art_ptr;
} Sdct_hdr;


#define DCT_HDR_SIZE 43


#include "libsdict.h"
/* Public interface to library */
/*
SDICT_API const char* sdict_get_library_version (void);
SDICT_API Sdict* sdict_new (int word_max, int art_max);
SDICT_API void  sdict_del (Sdict *sdict);
SDICT_API int sdict_load_dictionary (Sdict *sdict, const char* filename);
#ifdef _WIN32
SDICT_API int sdict_load_dictionary_w (Sdict* sdict, const wchar_t* filename);
#endif

#ifndef _WIN32
SDICT_API int sdict_load_dictionary_file (Sdict* sdict, FILE* file);
#endif

SDICT_API void sdict_unload_dictionary (Sdict *sdict);
SDICT_API int sdict_get_next_word(Sdict *sdic);
SDICT_API int sdict_get_prev_word(Sdict *sdic);
SDICT_API int sdict_get_article (Sdict *sdict);
SDICT_API int sdict_find_word (Sdict *sdict, utf8 *word);
SDICT_API int sdict_seek_word (Sdict *sdict, utf8 *word);
SDICT_API int sdict_rewind (Sdict *sdict);
#define SEEK_POS_MAX 1000
SDICT_API int sdict_seek_to (Sdict *sdict, int pos); // 0 - start, 1000 - end, 500 - 50%
SDICT_API int sdict_get_position (Sdict *sdict);
*/

#ifdef __LIBSDICT_C /* private */


static int sdict_load_dictionary_fd (Sdict* sdict, int fdes);
static Unit* read_unit (Sdict *sdict, int offset);
static uchar* string_from_unit(Unit *u);
static void del_unit(Unit *u);
static Sindex* read_sindex(Sdict *sdict, int offset);
static int parse_sindex(Sdict *sdict);
static int read_article_unit (Sdict *sdict, int offset);

#else /* __LIBSDICT_C public */


#endif /* __LIBSDICT_C  */




#ifdef __cplusplus
}
#endif /* __cplusplus */



#endif /* __LIBSDICT_INT_H_ */



/* __END__ */
