/*
 * $RCSfile: libunicode.c,v $
 * $Author: swaj $
 * $Revision: 1.6 $  
 */

/*
 * Was written by Alexey Semenoff (c) 2001-2006
 * See COPYING for copyiright policy
 */


#define _GNU_SOURCE

#ifdef _WIN32
#pragma warning(disable :4013)
#pragma warning(disable :4018)
#include "win32.h"
#endif


#include <stdio.h>
#include <string.h>
#include <errno.h>


#ifndef _WIN32
#include <stdint.h>
#endif



#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if defined(DEBUG) && !defined(DEBUG_LIBSDICT)
#undef DEBUG
#endif


#include "defines.h"
#include "debugging.h"
#include "util.h"


#define __LIBUNICODE_C 1
#include "libunicode.h"
#undef __LIBUNICODE_C

/*
 * strlen for utf8, returns number of chars
 */
int utf8_strlen (utf8* u)
{
	int i, tl;
	int len=0;
	uchar c;

	unless (u) {
		DBG("null arg");
		return 0;
	}

	tl = strlen(u);

	unless (tl) {
		DBG("zero len");
		return 0;
	}

	for (i=0; i<tl; i++) {
		c = u[i];

		if (c < 0x80) {
//			DBG("1 byte");
			len++;

		} else if ( (c >= 0xc0) && (c < 0xe0) ) {
			// 110xxxxx 10xxxxxx
//			DBG("2 bytes");
			len++;

		} else if ( (c >= 0xe0) && (c < 0xf0) ) {
			// 1110xxxx 10xxxxxx 10xxxxxx
			len++;

		} else if ( (c >= 0xf0 ) && (c < 0xf8) ) {
			// 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
			len++;

		} else if ( (c >= 0xf8 ) && (c< 0xfc) ) { 
			// 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			len++;

		} else if ( (c >= 0xfc ) && (c< 0xfe) ) { 
			// 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			len++;
		}
	}

	return len;
}


int utf8_strncmp (utf8* s, utf8* d, int l)
{
	int i; /*, rl; */
	uchar c;

	DBG("utf8_strncmp '%s' vs '%s'  l = '%d'", s, d, l);

	for (i=0; i < l; i++) {

		if ( (!*s) && (!*d) ) return RET_SUCCESS;

		if ( (!*s) || (!*d) ) return RET_FAILED;


		c = *s; 

		if (c != *d) return RET_FAILED;


		if ( (c >= 0xc0) && (c < 0xe0) ) {
			// 110xxxxx 10xxxxxx 2 bytes
			i--;

		} else if ( (c >= 0xe0) && (c < 0xf0) ) {
			// 1110xxxx 10xxxxxx 10xxxxxx
			i -= 2;

		} else if ( (c >= 0xf0 ) && (c < 0xf8) ) {
			// 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
			i -= 3;

		} else if ( (c >= 0xf8 ) && (c< 0xfc) ) { 
			// 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
                        i -= 4;

		} else if ( (c >= 0xfc ) && (c< 0xfe) ) { 
			// 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			i -= 5;
		}


		s++;
		d++;
	}


	return RET_SUCCESS; 
}


utf32* utf8_to_utf32 (utf8* u)
{
	utf8 *p1;
	utf32 *p2, *ptr;

	int len, ul;
	uchar c, c2, c3, c4, c5, c6;

	unless (u) {
		DBG("null arg");
		return NULL;
	}

	len = utf8_strlen(u);
	
	unless (len) {
		DBG("zero len");
		return NULL;
	}

	ul = strlen(u);

	ptr = smalloc ( (len + 1) * sizeof(utf32)); 
	
	unless (ptr) {
		DBG("malloc error");
		return NULL;
	}

	p1 = u;
	p2 = ptr;


	while (1) {

		if (p1 >= u + ul) break;

		c = *p1++;

		if (c < 0x80) {
			*p2 = (utf32)c;
			// DBG("1byte , 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xc0) && (c < 0xe0) ) { // 110xxxxx 10xxxxxx 
			if (p1 >= u + ul) break;

			c2 = *p1++;
			c  &= 0x1f;
			c2 &= 0x3f;
			*p2 = (c << 6) + c2;
			// DBG("2bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xe0) && (c < 0xf0) ) { // 1110xxxx 10xxxxxx 10xxxxxx
			if (p1 + 1 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;

			c  &= 0x0f;
			c2 &= 0x3f;
			c3 &= 0x3f;

			*p2 = (c << 12) + (c2 << 6) + c3;
			// DBG("3bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xf0 ) && (c < 0xf8) ) { // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 2 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;

			c  &= 0x07;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;

			*p2 =  (c << 18 ) + (c2 << 12) + (c3 << 6) + c4;
			// DBG("4bytes, 0x%x", *p2 );			
			p2++;


		} else if ( (c >= 0xf8 ) && (c< 0xfc) ) { // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 3 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;
			c5 = *p1++;

			c  &= 0x03;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;
			c5 &= 0x3f;

			*p2 =  (c << 24) + (c2 << 18 ) + (c3 << 12) + (c4 << 6) + c5;
			// DBG("5bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xfc ) && (c< 0xfe) ) { // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 4 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;
			c5 = *p1++;
			c6 = *p1++;

			c  &= 0x01;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;
			c5 &= 0x3f;
			c6 &= 0x3f;

			*p2 = (c << 30) + (c2 << 24) + (c3 << 18 ) + (c4 << 12) + (c5 << 6) + c6;
			// DBG("6bytes, 0x%x", *p2 );			
			p2++;


		} else {

			DBG("invalid char '0x%x'", (int)c); 
			*p2 = (utf32)REPLACE_CHAR;
			p2++;
		}
	}

	return ptr;
}


utf32* utf8_to_utf32_ninplace (utf8* u, utf32 *d, int n)
{
	utf8 *p1;
	utf32* p2; //, *ptr;
	int len, ul;
	uchar c, c2, c3, c4, c5, c6;

	unless (u && d)  {
		DBG("null arg");
		return 0;
	}

	len = utf8_strlen(u);
	
	unless (len) {
		DBG("zero len");
		return 0;
	}

	ul = strlen(u);

	p1 = u;
	p2 = d;

	while (n--) {
		if (p1 >= u + ul) break;

		c = *p1++;

		if (c < 0x80) {
			*p2 = (utf32)c;
			// DBG("1byte , 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xc0) && (c < 0xe0) ) { // 110xxxxx 10xxxxxx 
			if (p1 >= u + ul) break;

			c2 = *p1++;
			c  &= 0x1f;
			c2 &= 0x3f;
			*p2 = (c << 6) + c2;
			// DBG("2bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xe0) && (c < 0xf0) ) { // 1110xxxx 10xxxxxx 10xxxxxx
			if (p1 + 1 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;

			c  &= 0x0f;
			c2 &= 0x3f;
			c3 &= 0x3f;

			*p2 = (c << 12) + (c2 << 6) + c3;
			// DBG("3bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xf0 ) && (c < 0xf8) ) { // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 2 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;

			c  &= 0x07;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;

			*p2 =  (c << 18 ) + (c2 << 12) + (c3 << 6) + c4;
			// DBG("4bytes, 0x%x", *p2 );			
			p2++;


		} else if ( (c >= 0xf8 ) && (c< 0xfc) ) { // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 3 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;
			c5 = *p1++;

			c  &= 0x03;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;
			c5 &= 0x3f;

			*p2 =  (c << 24) + (c2 << 18 ) + (c3 << 12) + (c4 << 6) + c5;
			// DBG("5bytes, 0x%x", *p2 );			
			p2++;

		} else if ( (c >= 0xfc ) && (c< 0xfe) ) { // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
			if (p1 + 4 >= u + ul) break;

			c2 = *p1++;
			c3 = *p1++;
			c4 = *p1++;
			c5 = *p1++;
			c6 = *p1++;

			c  &= 0x01;
			c2 &= 0x3f;
			c3 &= 0x3f;
			c4 &= 0x3f;
			c5 &= 0x3f;
			c6 &= 0x3f;

			*p2 = (c << 30) + (c2 << 24) + (c3 << 18 ) + (c4 << 12) + (c5 << 6) + c6;
			// DBG("6bytes, 0x%x", *p2 );			
			p2++;


		} else {

			DBG("invalid char '0x%x'", (int)c); 
			*p2 = (utf32)REPLACE_CHAR;
			p2++;
		}
	}

	return d;
}


int utf32_strlen (utf32* u)
{
	int len=0;

	unless (u) {
		DBG("null arg");
		return 0;
	}

	while (*u++) {
		len++;
	}

	return len;
}


int utf32_strnlen (utf32* u, int l)
{
	int i;
	int len=0;

	for (i=0; i<l; i++) {
		unless (*u++) return len;
		len++;
	}
	return len;

}

int utf32_strncmp (utf32* s, utf32* d, int l)
{
	int i;

	for (i=0; i<l; i++) {
		if (*s != *d) return RET_FAILED;
		s++;
		d++;
	}

	return RET_SUCCESS;
}




	
/* __END__ */






