/*
 * $RCSfile: sdict-main.c,v $
 * $Author: swaj $
 * $Revision: 1.3 $  
 */

/*
 * Was written by Alexey Semenoff (c) 2001-2006
 * See COPYING for copyiright policy
 */


#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "defines.h"
#include "debugging.h"

#include "libunicode.h"
#include "libsdict.h"

#define __SDICT_MAIN_C 1
#include "sdict-main.h"
#undef __SDICT_MAIN_C


static
void term_init (void)
{
	if (tcgetattr(fileno(stdin),&work_tio)) {
		fprintf (OUTPUT,"Unable to get terminal attributes\n");
		exit (1);
	}

	saved_tio = work_tio;

	work_tio.c_lflag &= (~ICANON & ~ECHO);
	work_tio.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL | ECHOCTL | ECHOPRT | ECHOKE);

	if (tcsetattr(fileno(stdin), TCSANOW, &work_tio)) {
		fprintf (OUTPUT,"Unable to set terminal attributes\n");
		exit (1);
	}
}


static
void term_cleanup(void)
{
	if (tcsetattr(fileno(stdin), TCSANOW, &saved_tio)) {
		fprintf (OUTPUT,"Unable to set terminal attributes\n");
	}
}


static
int is_buffer_cmd(void)
{

	return  input_string[0] == '/' ? RET_SUCCESS : RET_FAILED;
}


static
uchar* get_char_mapping (uchar c)
{
	KeyMapping *ptr;


	unless (which_km) {
		DBG("null table");
		return NULL;
	}

	if (is_buffer_cmd()==RET_SUCCESS) {
		ptr = la;
	} else {
		ptr = which_km;
	}

	while (ptr->keyb_c) {
		if (ptr->keyb_c == c) {
			if (strlen (ptr->mapping)) {
				return ptr->mapping;
			}
		}
		
		ptr++;
	}

	return NULL;
}


static
int get_command(void)
{
	int i;


	if (is_buffer_cmd()==RET_FAILED)
		return C_ILL;


	for (i=0; i<(sizeof(cmds)/sizeof(Commands)); i++) {

                if (cmds[i].command_no==C_ILL) {
			return C_ILL;
		}

                unless (
			strncmp(
				cmds[i].command_s,
				&input_string[1],
				strlen(cmds[i].command_s)
			) ){

			return cmds[i].command_no;
		}
        }
	return C_ILL;
}


static
void print_intro(void)
{
	printf ("\n");
	printf ("Sdict console client started. Version %s.\n", VERSION);
	printf ("This program was written by Alexey Semenoff (c) 2001-2006. All rights reserved.\n");
	printf ("See COPYING for copyiright policy.\n");
	printf ("\n");
	printf ("Type '/help' to get usage information\n");

	printf ("\n");
}


static
void print_help(void)
{
	printf ("\n");

	printf ("Control commands are started with '/' and listed below:\n\n");
	printf ("/help                      This help\n");
	printf ("/quit     /q               Quit program\n");
	printf ("/kb la    /kl              Use latin keyboard layout\n");
	printf ("/kb ru    /kr              Use russian koi8 keyboard layout\n");
	printf ("/load     /l    filename   Load .dct file\n");
	printf ("/info     /i               Dictionary info\n");
	printf ("/!cmd                      Run shell command i.e. /!ls\n");
	printf ("\n");

	printf ("Otherwise enter word for translation and press\n");
	printf ("  <Tab>    see nearest words in dictionary\n");
	printf ("  <Enter>  translate the word\n");
	printf ("\n");
}


static
void print_cmd_ok(void)
{
	printf ("Ok\n");
}


static
void load_dct_file_err(void)
{
	printf ("Usage /load filename\n");
}


static
void print_dct_not_loaded(void)
{
	printf ("\nDictionary is not loaded\n");
}


static
void print_dct_info(void)
{

	unless (dct_loaded) {
		print_dct_not_loaded();
	} else {
		printf ("\nDictionary information: \n");

		printf ("Title '%s'\n", sdict_get_dic_title (sdict) );
		printf ("Copyright '%s'\n",  sdict_get_dic_copyright (sdict) );
		printf ("Version '%s'\n", sdict_get_dic_version (sdict) );

		printf ("Languages %s/%s\n",
			sdict_get_dic_word_lang (sdict),
			sdict_get_dic_article_lang (sdict) );
		printf ("Words '%d'\n\n", sdict_get_dic_words (sdict) );
	}
}


static
void load_dct_file(void)
{
	char *ptr;


	ptr = strchr(input_string, ' ');

	unless (ptr) {
		load_dct_file_err();
		return;
	}

	ptr++;

	unless (strlen (ptr)) {
		load_dct_file_err();
		return;
	}

	printf ("Loading file '%s'\n", ptr);

	if (dct_loaded)
		sdict_unload_dictionary(sdict);


	if (sdict_load_dictionary (sdict, ptr) != RET_SUCCESS) {
                printf ("Unable load dictionary: '%s'\n", sdict_get_strerror(sdict) );
		return;
        }
	dct_loaded = 1;

	printf ("\nLoaded dictionary '%s'\n", ptr);

	print_dct_info();
}


static
void find_word(void)
{
	unless (dct_loaded) {
		print_dct_not_loaded();
		return;
	}

	if (sdict_find_word(sdict, input_string)==RET_FAILED) {
		printf ("\nNot found\n");
		return;
	}

	printf ("\n");
	printf ("%s\n\n", sdict_get_last_word (sdict) );
	printf ("%s\n", sdict_get_last_article (sdict) );
	printf ("\n");
}


static
void make_completion(void)
{
	int i;


	if (is_buffer_cmd()==RET_SUCCESS) {
		/* command completion, not implemented */

		switch (get_command()) {

		case C_LOAD:
			printf ("\n");
			system ("ls *.dct");
			printf ("%s%s", prompt, input_string);
			break;

		default:
			break;
		}


		return;
	}

	unless (dct_loaded) {
		return;
	}

	if (sdict_seek_word(sdict, input_string)==RET_FAILED) {
		return;
	}

#define WORD_COMPL_ITEMS 20

	printf ("\n");

	for (i=0; i<WORD_COMPL_ITEMS; i++) {

		if (sdict_get_next_word(sdict)==RET_FAILED) return;

		printf ("\n%s",  sdict_get_last_word (sdict) );
        }

	printf ("\n\n%s%s", prompt, input_string);

}


static
void run_shell_cmd(void)
{
	char *ptr =input_string + 2;  

	if (strlen (ptr)) {

		term_cleanup();
		system (ptr);
		term_init ();
	}
}


static
void run_cmd(void)
{
	int co;

	co = get_command();

	/* DBG("'%s' %d", input_string, co); */

	switch (co) {

	case C_QUIT:
		term_cleanup();
		exit (0);

	case C_HELP:
		print_help();
		break;


	case C_KBLA:
		which_km = fi;
		print_cmd_ok();
		break;

	case C_KBRU:
		which_km = ru;
		print_cmd_ok();
		break;

	case C_LOAD:
		load_dct_file();
		break;

	case C_INFO:
		print_dct_info();
		break;

	case C_EXT:
		run_shell_cmd();
		break;

	case C_ILL:
	default:
		printf ("Syntax error, try /help\n");
		break;

	}
}


static
uchar *get_input_string(void)
{
	uchar c, *p;

	memset (input_string, 0x0, sizeof(input_string));
	
	printf ("%s", prompt);

	while (1) {
  
		c = getchar();
		/* DBG("c = '0x%x'\n", c); */

		switch (c) {

		case '\n':
		case '\r':
			printf ("\n");
			unless (input_string_pos)
				return NULL;

			input_string[input_string_pos] = 0x0;
			input_string_pos = 0;
			return input_string;

		case BACKSPACE:
			if (input_string_pos) {
				printf ("%s", "\b \b");
				if (input_string_pos) {
					input_string_pos--;
				}

				/* workaround for latin / cyrillic - 2 byte utf */
				if ( (input_string[input_string_pos] >= 0x80) || 
				     (input_string[input_string_pos+1] >= 0x80) )  {
				     input_string_pos--;
				}

				input_string[input_string_pos]= 0x0;
			}
			continue;

		case TABULATION:
			if (input_string_pos) {
				make_completion();
			}
			break;

		default:
			if (input_string_pos >= INPUT_STRING_LEN) {
				printf ("%c", '\a');

			} else {
				p = get_char_mapping(c);

				if (p) {
					printf ("%s", p);
					snprintf (input_string + strlen(input_string),
						  sizeof(input_string) - strlen(input_string),
						  "%s", p);
					
					input_string_pos += strlen(p);
					if (input_string_pos >= INPUT_STRING_LEN)
						input_string_pos = INPUT_STRING_LEN;

				}
			}
			continue;
		}
	}
}


/****
 *  *
 ****/

int main (void)
{
	uchar *s;


	sdict = sdict_new (0,0);

        unless (sdict) {
                printf ("unable create sdict object, fatal\n");
                exit (1);
        }

	term_init();

	SET_PROMPT(DEFAULT_PROMPT);

	print_intro();

	while (1) {
		s = get_input_string();

		if (s && strlen(s)) {
			if (is_buffer_cmd() == RET_SUCCESS) {
				run_cmd();
			} else {
				find_word();
			}
		}
	}

	exit (0);
}



/* __END__ */
